"""Top-level package for interval_search."""

__author__ = """Matthew Andres Moreno"""
__email__ = 'm.more500@gmail.com'
__version__ = '0.2.0'

from .binary_search import binary_search
from .doubling_search import doubling_search
from .interval_search import interval_search

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    'binary_search',
    'doubling_search',
    'interval_search',
]



#!/usr/bin/env python

'''
`binary_search` tests for `interval_search` package.
'''

import interval_search as inch


def test_binary_search_singleton():
    assert inch.binary_search(lambda __: True, 10, 10) == 10


def test_binary_search():
    assert inch.binary_search(lambda x: x >= 5, 0, 100) == 5


def test_fruitless_binary_search():
    assert inch.binary_search(lambda x: False, 0, 100) is None

def test_empty_binary_search():
    assert inch.binary_search(lambda x: False, 100, 99) is None
    assert inch.binary_search(lambda x: False, 100, 0) is None
    assert inch.binary_search(lambda x: True, 100, 99) is None
    assert inch.binary_search(lambda x: True, 100, 0) is None



import typing


def binary_search(
    predicate: typing.Callable[[int], bool],
    lower_bound: int,
    upper_bound: int,
) -> typing.Optional[int]:
    """Find the positive integer threshold below which a search criteria is
    never satisfied and above which it is always satisfied.

    Parameters
    ----------
    predicate : callable object
        Returns whether an integer value satisfies the search criteria.
    lower_bound : int
        Lower bound for the binary search, inclusive.
    upper_bound : int
        Upper bound for the binary search, inclusive.

    Returns
    -------
    guess
        The lowest integer value that satisfies the search criteria, and None
        if upper_bound does not satisfy the search criteria or search range is
        empty (i.e., lower_bound > upper_bound).
    """

    if lower_bound > upper_bound:
        return None
    if lower_bound == upper_bound:
        if predicate(lower_bound):
            return lower_bound
        else:
            return None

    midpoint = (lower_bound + upper_bound) // 2

    if predicate(midpoint):
        return binary_search(predicate, lower_bound, midpoint)
    else:
        return binary_search(predicate, midpoint + 1, upper_bound)



import typing

from .binary_search import binary_search


def doubling_search(
    predicate: typing.Callable[[int], bool],
    lower_bound: int = 0,
) -> int:
    """Find the positive integer threshold below which a search criteria is
    never satisfied and above which it is always satisfied.

    Parameters
    ----------
    predicate : callable object
        Returns whether an integer value satisfies the search criteria.
    lower_bound : int, optional
        The initial guess. Should be less than or equal to the first value that
        satsfies the search criteria. Used for recursion. Default is 0.

    Returns
    -------
    threshold
        The lowest integer value that satisfies the search criteria.
    """

    assert lower_bound >= 0, lower_bound

    bound = 1
    while not predicate(lower_bound + bound):
        bound *= 2

    prev_bound = bound // 2
    prev_guess = lower_bound + prev_bound
    return binary_search(predicate, prev_guess, lower_bound + bound)
